#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

int reservedSectors = 64, useInvertedData = 1;

int RS_InitComm(int comPort);
int RS_CloseComm(void);
int RS_SetCommSpeed(long speed);
int RS_SendData(char *data, int len);
int RS_ReceiveData(char *data, int len);
void RS_Wait(void);
void RS_Drain(void);

/***************************************************************************/
int RS_InitComm(int comPort);
int RS_CloseComm(void);
int RS_SetCommSpeed(long speed);
int RS_SendData(char *data, int len);
int RS_ReceiveData(char *data, int len);
void RS_Wait(void);
void RS_Drain(void);

/*

FILE: rs232.c - serial connection functions
DESCRIPTION: Serial communication for unix / DOS

*/
#define BAUDRATE B9600
#define DEF_BAUDRATE 9600

unsigned int baudRate=DEF_BAUDRATE;
volatile int * volatile breakPtr = NULL;

void RS_SetBreakIntVar(volatile int *breakVar) {
    breakPtr = breakVar;
}

#ifndef WIN32

#include <sys/types.h>
#include <termios.h>
#include <fcntl.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* FreeBSD termios-based serial communication */

int newsfd;
struct termios oldtio;

void RS_Wait(void) {
    usleep(100000);
}

void RS_Drain(void) {
    tcdrain(newsfd);
}

int RS_InitComm(int comPort) {
    static char commDevice[80];
#if !defined(__linux__) && !defined(linux)
    sprintf(commDevice, "/dev/cuaa%d", comPort-1);
#else
    sprintf(commDevice, "/dev/ttyS%d", comPort-1);
#endif
    /* support ttyUSBxx */
    if (comPort >= 1000) {
	sprintf(commDevice, "/dev/ttyUSB%d", comPort - 1000);
    }
    newsfd = open(commDevice, O_RDWR | O_NOCTTY);
    if (newsfd <0) {
	perror(commDevice);
	return -1;
    }
    tcgetattr(newsfd,&oldtio); /* save current port settings */
    return RS_SetCommSpeed(baudRate);
}

void RS_SendBreak() {
    tcsendbreak(newsfd, 0); /* 0.25 .. 0.5 sec break */
}

int RS_SendData(char *data, int len) {
    if (write(newsfd, data, len) < 0)
	perror("Write error");
    return 0;
}

int RS_ReceiveData(char *data, int len) {
  int i = 0;

  while (len - i > 0) {
      int k = read(newsfd, data+i, len - i);

//fprintf(stderr, "Read %d\n", k);
      if (k < 0) {
	  /* read error, restart read */
//fprintf(stderr, "Read err, restart\n");
      } else {
	  i += k;
	  /* timeout, check breakPtr */
	  if (k == 0 && breakPtr) {
	      if (breakPtr && *breakPtr) {
//fprintf(stderr, "timeout + break\n");
		  return -(i+1);
	      }
	  }
      }
  }
  return 0;
}

int RS_CloseComm(void) {
    /* restore old port settings */
    tcsetattr(newsfd,TCSANOW,&oldtio);
    /* close serial port */
    close(newsfd);
    return 0;
}

#if defined(__linux__) || defined(linux)
#include <signal.h>
#include <sys/resource.h>
#include <linux/serial.h>
#include <sys/ioctl.h>
#endif



int Set_DTR(void) {
    int nRet;
    int status;
    status = 0;
    nRet = ioctl(newsfd, TIOCMGET,&status);
    if (nRet < 0)
	return 0;
    status |= TIOCM_DTR;
    nRet = ioctl(newsfd, TIOCMSET, &status);
    if (nRet < 0)
	return 0;
    return 1;
}


int Clr_DTR(void) {
    int nRet;
    int status;
    status = 0;
    nRet = ioctl(newsfd, TIOCMGET,&status);
    if (nRet < 0)
	return 0;
    status &= ~TIOCM_DTR;
    nRet = ioctl(newsfd, TIOCMSET, &status);
    if (nRet < 0)
	return 0;
    return 1;
}


struct Speed {
  unsigned long symbol;
  int bps;
} speeds[] = {
  {B0      , 	  0},	/* hang up */
  {B50     ,      50},
  {B75     ,      75},
  {B110    ,     110},
  {B134    ,     134},
  {B150    ,     150},
  {B200    ,     200},
  {B300    ,     300},
  {B600    ,     600},
  {B1200   ,    1200},
  {B1800   ,    1800},
  {B2400   ,    2400},
  {B4800   ,    4800},
  {B9600   ,    9600},
  {B19200  ,   19200},
  {B38400  ,   38400},
  {B57600  ,   57600},
  {B115200 ,  115200},
  {B230400 ,  230400},
#ifdef B460800
  {B460800 ,  460800},
#endif
#ifdef B500000
  {B500000 ,  500000},
#endif
#ifdef B576000
  {B576000 ,  576000},
#endif
#ifdef B921600
  {B921600 ,  921600},
#endif
#ifdef B1000000
  {B1000000, 1000000},
#endif
#ifdef B1152000
  {B1152000, 1152000},
#endif
#ifdef B1500000
  {B1500000, 1500000},
#endif
#ifdef B2000000
  {B2000000, 2000000},
#endif
#ifdef B2500000
  {B2500000, 2500000},
#endif
#ifdef B3000000
  {B3000000, 3000000},
#endif
#ifdef B3500000
  {B3500000, 3500000},
#endif
#ifdef B4000000
  {B4000000, 4000000},
#endif
};

/*

FUNCTION: SetCommSpeed()
DESCRIPTION: Sets communications speed
INPUTS: speed - desired speed in bps
RESULT: 0 if there were no errors, -1 otherwise
SEE ALSO:

*/
int RS_SetCommSpeed(long speed) {
    struct termios speedtio;
    unsigned long speedSymbol;
    int i = 0;
#if defined(__linux__) || defined(linux)
    int customSpeed = 0;
#endif

    while (i < sizeof(speeds)/sizeof(speeds[0]) && speed != speeds[i].bps)
	i++;

    if (i >= sizeof(speeds)/sizeof(speeds[0])) {
#if defined(__linux__) || defined(linux)
	customSpeed = 1;
	speedSymbol = B38400;
#else
	speedSymbol = speed; /*FreeBSD takes a value instead of bit mask!*/
#endif
    } else {
	speedSymbol = speeds[i].symbol;
    }

    /* set new port settings for canonical input processing */
    memset(&speedtio, 0, sizeof(speedtio));
    tcgetattr(newsfd,&speedtio); /* save current port settings */
    speedtio.c_cflag = CS8 | CLOCAL | CREAD ;
    speedtio.c_iflag = IGNPAR;
    speedtio.c_oflag = 0;
    speedtio.c_lflag = 0;

    cfmakeraw(&speedtio);

    speedtio.c_cc[VMIN] = 0;
    speedtio.c_cc[VTIME] = 30;//15; /* in 0.1sec granularity, VMIN=0 */

#if defined(__linux__) || defined(linux)
    if (customSpeed) {
	struct serial_struct ss;
	int closestSpeed;

	ioctl(newsfd, TIOCGSERIAL, &ss);
	ss.flags = (ss.flags & ~ASYNC_SPD_MASK) | ASYNC_SPD_CUST;
	ss.custom_divisor = (ss.baud_base + (speed / 2)) / speed;
	if (ss.custom_divisor == 0)
	    ss.custom_divisor = 1;
	closestSpeed = ss.baud_base / ss.custom_divisor;

	if (closestSpeed < speed * 98 / 100 ||
	    closestSpeed > speed * 102 / 100) {
	    fprintf(stderr, "Cannot set serial port speed to %d. "
		    "Closest possible is %d\n",
		    (int)speed, (int)closestSpeed);
	}
	ioctl(newsfd, TIOCSSERIAL, &ss);
    }
#endif /*defined(__linux__) || defined(linux)*/

    cfsetispeed(&speedtio, speedSymbol);
    cfsetospeed(&speedtio, speedSymbol);
    if (tcsetattr(newsfd, TCSADRAIN, &speedtio)==-1) {
	printf("Serial port input speed (%ld) setting failed, errno: %d\n",
	       speed, errno);
	return -1;
    }
//   Set_DTR();
    Clr_DTR();
    return 0;
}

#else

/* See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/about_communications_resources.asp */
#include <windows.h>

/* module variables*/
HANDLE       SerialHandle;
DCB          Control;	      /* device control block */
COMMTIMEOUTS TimeOuts;

int RS_InitComm(int comPort)  {
    static char portName[80];
    if (comPort > 9) {
	sprintf(portName, "\\\\.\\COM%d", comPort);
    } else {
	sprintf(portName, "COM%d", comPort);
    }
    SerialHandle = CreateFile(portName,
			      GENERIC_READ|GENERIC_WRITE,
			      0,                  /* not shared */
			      NULL,		 /* no security attributes*/
			      OPEN_EXISTING,
			      FILE_ATTRIBUTE_NORMAL|FILE_FLAG_NO_BUFFERING|
			      FILE_FLAG_WRITE_THROUGH,
			      NULL);
  
    if (SerialHandle == INVALID_HANDLE_VALUE) {
	fprintf(stderr, "Couldn't open serial port!\n");
	return -1;
    }
    if (!BuildCommDCB("baud=1200 parity=N data=8 stop=1", &Control)) {
	fprintf(stderr, "Couldn't build control block!\n");
	return -1;
    }
    Control.DCBlength = sizeof(DCB);
    Control.BaudRate = baudRate;
    Control.fBinary = TRUE;
    Control.fParity = FALSE;
    Control.fOutxCtsFlow = FALSE;
    Control.fOutxDsrFlow = FALSE;
    Control.fDtrControl = DTR_CONTROL_DISABLE;
    Control.fDsrSensitivity = FALSE;
    Control.fTXContinueOnXoff = FALSE;
    Control.fOutX = FALSE;
    Control.fInX = FALSE;
    Control.fErrorChar = FALSE;
    Control.fNull = FALSE;
    Control.fRtsControl = RTS_CONTROL_DISABLE;
    Control.fAbortOnError = TRUE;
    /* set user defined values */
    if (!SetCommState(SerialHandle, &Control)) {
	fprintf(stderr, "Couldn't set control state!\n");
	return -1;
    }

    TimeOuts.ReadIntervalTimeout    = 0; /* max. time between two bytes in ms*/
    TimeOuts.ReadTotalTimeoutMultiplier = 0; /*1second/byte allowed*/
    TimeOuts.ReadTotalTimeoutConstant   = 4000; /*ms overhead in the start of tx*/
    TimeOuts.WriteTotalTimeoutMultiplier= 0; /*ms/write allowed */
    TimeOuts.WriteTotalTimeoutConstant  = 0; /*ms overhead in the start of tx*/

    if (!SetCommTimeouts(SerialHandle, &TimeOuts)) {
	fprintf(stderr, "Couldn't set timeout values!\n");
	return -1;
    }
    /* clear input buffer */
    PurgeComm(SerialHandle, PURGE_RXABORT | PURGE_RXCLEAR);
    return 0;
}

int RS_ReceiveData(char *data, int len) {
    int i = 0;

    while (len - i > 0) {
	DWORD read = 0;

	ReadFile(SerialHandle, (BYTE *)data+i, len-i, &read, NULL);
	if (read < 0) {
	    /* read error */
	} else {
	    i += read;
	    /* timeout, check breakPtr */
	    if (read == 0 && breakPtr) {
		if (breakPtr && *breakPtr)
		    return -(i+1);
	    }
	}
    }
    return 0;
}

void RS_SendBreak() {
    /*tcsendbreak(newsfd, 0);*/ /* 0.25 .. 0.5 sec break */
    RS_Drain();
    SetCommBreak(SerialHandle); /*EscapeCommFunction(SerialHandle, SETBREAK);*/
    Sleep(500);
    ClearCommBreak(SerialHandle); /*EscapeCommFunction(SerialHandle, CLRBREAK);*/
}

int RS_SendData(char *data, int len) {
    DWORD written;
  
    WriteFile(SerialHandle, (BYTE *)data, len, &written, NULL);
    if (written != len) {
	fprintf(stderr, "Write byte failed!\n");
	return -1;
    }
	return 0;
}

int RS_CloseComm(void) {
    PurgeComm(SerialHandle, PURGE_RXABORT | PURGE_RXCLEAR);
    CloseHandle(SerialHandle);
    return 0; 
} 
 
int RS_SetCommSpeed(long speed) {
    if (!GetCommState(SerialHandle, &Control)) {
	fprintf(stderr, "Couldn't get control state!\n");
	return -1;
    }
    Control.BaudRate = speed;
    if (!SetCommState(SerialHandle, &Control)) {
	fprintf(stderr, "Couldn't set control state!\n");
	return -1;
    }
    return 0;
}

void RS_Drain(void) {
    FlushFileBuffers(SerialHandle);
}

void RS_Wait(void) {
    Sleep(300);
}
#endif
/***************************************************************************/


#define BLSIZE (512) //512 bytes
int fileSize = 0;
int blSize = -1;
char block[BLSIZE+10];

void GetBlock(FILE *fp) {
    memset(block+6, 0, BLSIZE);
    blSize = fread(block+6, 1, BLSIZE, fp);
    fileSize += blSize;
    if (blSize & 1)
	blSize++;
}
/*
  byte 0 - 0xbb - start marker
  byte 1 - size EaaaaaaS - a = extra address bits, E=end of program, S=end of file
  bytes 2&3 - 16-bit block number
  bytes 4&5 - inverse of data checksum
  bytes 6..517 - data
 */

void SendBlock(long addr, int size /*1=ok,0=eof,0x81=eop*/) {
    int i, ch = 0xffff, extra = (addr >> 16);
    block[0] = 0xbb; /*Start marker*/
    block[1] = (size & 0x81) | ((addr>>15)&0x7e); /*Size in blocks, extra block address bits*/
    block[2] = addr>>8;
    block[3] = addr;


    if (useInvertedData && addr >= reservedSectors) {
	for (i=0;i<512;i++) {
	    block[6+i] ^= 0xff;
	}
    }
    for (i=0;i<512/2;i++) {
	ch += (block[6+2*i]<<8) + (block[7+2*i]&0xff);
    }
    ch = ~ch;
//printf("%04x\n", ch &0xffff);
    block[4] = ch>>8;
    block[5] = ch;
    RS_SendData(block, 512+6);
}


/*************************************************************************/
/* SBFS - Filesystem for the UART Voicemail Firmware                     */
/*************************************************************************/
#define SBFS_COOKIE (('S'<<8)+('B'))
#define GET_FILE_NUM (0xffffu)
#define FREE_BLOCK (0xffffu)
#define FIND_FREE (0xfffeu)
#define FILE_LIST (0xfffdu)
#define FIND_NAME (0xfffcu)
/// Files in an SBFS disk are stored in chunks.
/// a chunk must start at an eraseblock boundary.
/// Chunks must be stored in correct order, e.g. first chunk of a file is in the lowest address.
/// Chunks are to be continuous and as large as possible.
/// There is no central directory: to find all files, all eraseblock beginnings in disk free space must be read.
#if 0
typedef struct SbfsChunkHeaderStruct {
    u_int16 sbfsCookie; //'SB'
    u_int16 length; //Number of data blocks in this Chunk
    u_int16 chunkType; //0xffff = unused, 0x0000 = data, 0x0001 = Start of file, 0x0002 = end of file, 0x0003 = Start and End of file
    u_int16 chunkSubType; //Unused
    u_int32 bytesUsed; //Number of bytes stored in the file by the end of chunk.
    u_int32 nextLba; //Location of next block of this file
    u_int32 eraseBlockSize; //Size of allocation block in lbas. This field helps to read a filesystem written by someone else.
    u_int16 padding[6];//for future extensions
    u_int16 fileName[128-16];//Name of the file in this block. 
    //This header's size is 256 bytes. The second 256 bytes in the header block can extend the file name and/or contain other info.
} SbfsChunkHeader;
#endif

#define UNUSED (1<<15)
#define ERRONEOUS (1<<14)
#define START_OF_FILE (1<<0)
#define END_OF_FILE (1<<1)

int AddVoiceFile(FILE *fp, const char *voiceFile) {
    FILE *in = fopen(voiceFile, "rb");
    if (in) {
	unsigned char blk[512];
	long szBytes, len, blks = 0;
	fseek(in, 0, SEEK_END);
	szBytes = ftell(in);
	fseek(in, 0, SEEK_SET);
	memset(blk, -1, sizeof(blk));

	blk[0] = 'S';
	blk[1] = 'B';
	len = (szBytes + 511) / 512; /* Does not include 4kB alignment */
	if (len > 65535) {
	  fprintf(stderr, "Error: Cannot add a %ld-block file to %s\n",
		  len, voiceFile);
	  return 1;
	}
	if (len > 32767) {
	    fprintf(stderr, "Warning: Adding a %ld-block file to %s\n",
		    len, voiceFile);
	}
	/* Note: max size is 65535*512=31.999MB, 32767*512=15.999MB preferred.*/
	blk[2] = len >> 8;
	blk[3] = len;
	blk[4] = 0; blk[5] = 3; /* Both start and end of file */
	blk[6] = 0; blk[7] = 0; /* unused */
	blk[8] = szBytes>>8;
	blk[9] = szBytes;
	blk[10] = szBytes>>24;
	blk[11] = szBytes>>16;
	blk[12] = blk[13] = blk[14] = blk[15] = 0; /* nextLba */
	blk[16] = blk[17] = blk[18] = blk[19] = 0; /* eraseBlockSize - decorative */
	/* 20/21, 22/23, 24/25, 26/27, 28/29, 30/31 are padding */
	{
	    const char *t = voiceFile;
	    char *t2;
	    int i = 0;
	    while ((t2 = strchr(t, '/'))) {
		t = t2 + 1;
	    }
	    do {
		blk[32+i*2] = 0;
		blk[33+i*2] = *t; /* One char per word */
		i++;
	    } while (*t++ && i < (128-16));
	}
	fwrite(blk, 1, sizeof(blk), fp);
	blks = 1;

	while (fread(blk, 1, sizeof(blk), in)) {
	    fwrite(blk, 1, sizeof(blk), fp);
	    blks++;
	}
	fclose(in);
	if (blks & 7) {
	    fprintf(stderr, "Adding padding %d to %s\n",
		    (int)(8-(blks&7)), voiceFile);
	    memset(blk, -1, sizeof(blk));
	    while ((blks & 7)) {
		fwrite(blk, 1, sizeof(blk), fp);
		blks++;
	    }
	}
	return 0;
    } else {
	return 1;
    }
}



/*************************************************************************/


#include "autoprogimg.h" /* VS1000 programmer */

void putch(int c) {
    unsigned char ch = c;
    RS_SendData(&ch, 1);
}
void putword(unsigned short w) {
    unsigned char ch[2];
    ch[0] = w>>8;
    ch[1] = w;
    RS_SendData(ch, 2);
}



long sentblocks = 0;
int SendFile(const char *name, int startBlock, int cont, int last) {
    FILE *fp;
    long addr = startBlock-1;
    unsigned char o = 0;

//fprintf(stderr, "\ncont %d\n", cont);
    fp = name ? fopen(name, "rb") : NULL;
    if (!fp) {
	printf("Could not open '%s' for reading\n", name ? name : "");
	exit(EXIT_FAILURE);
    }

    while (1) {
	unsigned char d;
	int c;

//fprintf(stderr, "Continuing with another file..\n");
	if (cont) {
	    o = ('' & 255);
	    d = 0x42;
	    blSize = -1;
	    goto doit;
	}
	if (!RS_ReceiveData((char *)&d, 1)) {
	doit:
	    cont = 0;
	    c = d;
//fprintf(stderr, "%c", isprint(c) ? c : ' ');
	    if (o != ('' & 255)) {
		//printf("old %d, old<-%d %d\n", o, d, '');
		o = d;
		continue;
	    }
	    o = d;
	    //printf(" old <- %d, c = %d\n", o, c);
	    if (c == 0x42) {
		if (!blSize) {
		    fclose(fp);
		    fp = NULL;
		    break;
		}
		fprintf(stderr, "\r%4d bytes block %ld", fileSize, addr);
		addr++;
		GetBlock(fp);
		SendBlock(addr, blSize ? 1 : (last ? 255 : 0));
		sentblocks++;
	    } else if (c == 0x43) {
		fprintf(stderr, "\nUART Transfer Failed, Retrying..\n");
		SendBlock(addr, blSize ? 1 : (last ? 255 : 0));
	    } else if (c == 0x44) {
		fprintf(stderr, "\nSPI Programming Failed, Retrying..\n");
		return 1;
	    } else if (c == 0x45) {
		fprintf(stderr, "\nSPI Programming Successful!\n");
		return 0;
	    } else if (c == 0x40) {
		fprintf(stderr, "VS1000 Ready, performing chip erase.\n");
	    } else if (c == 0x14) {
		fprintf(stderr, "2MB\n");
	    } else if (c == 0x15) {
		fprintf(stderr, "4MB\n");
	    } else if (c == 0x16) {
		fprintf(stderr, "8MB\n");
	    } else if (c == 0x17) {
		fprintf(stderr, "16MB\n");
	    } else if (c == 0x18) {
		fprintf(stderr, "32MB\n");
	    } else if (c == 0x19) {
	        fprintf(stderr, "64MB\n"); //Cypress S25FL512SDPMFIG11
	    } else if (c == 0x41) {
		fprintf(stderr, "Chip Erase complete.\n");
	    } else {
		fprintf(stderr, "Unknown response %c (0x%02x)\n",
			isprint(c) ? c : ' ', c);
	    }
	} else {
	    /* Timeout -- should wait until ~30 sec. */
	    //fprintf(stderr, "\nTimeout before starting!\n");
	    //return 1;
	}
    }
    fprintf(stderr, "\n");
    return 0;
}


#define MAX_VOICEFILES 128
#include <time.h>
int main(int argc, char *argv[]) {
    FILE *fp = NULL;
    char *inName = NULL, *fatName = NULL;
    const char *tmpName = NULL;
    const char *voiceFile[MAX_VOICEFILES] = {NULL};
    int port = 1, n, voiceFiles = 0, err = 0;
    time_t stime = time(NULL), etime;

    baudRate = 115200;
    for (n=1; n<argc; n++) {
	if (!strncmp(argv[n], "-r", 2) && argv[n][2]) {
	    reservedSectors = strtol(argv[n]+2, NULL, 0);
	} else if (!strncmp(argv[n], "-r", 2) && n+1<argc) {
	    reservedSectors = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-p") && n+1<argc) {
	    port = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-s") && n+1<argc) {
	    baudRate = strtol(argv[++n], NULL, 0);
	} else if (!strcmp(argv[n], "-noinvert")) {
	    useInvertedData = 0;
	} else if (!strcmp(argv[n], "-f") && n+1<argc) {
	    if (voiceFiles < MAX_VOICEFILES) {
		voiceFile[voiceFiles++] = argv[++n];
		reservedSectors = 32;
		useInvertedData = 0;
	    } else {
		err = 1;
	    }
	} else if (!strcmp(argv[n], "-l") && n+1<argc) {
	    if (inName) {
		fprintf(stderr, "Warning: only one input file allowed, "
			"%s discarded\n", inName);
	    }
	    inName = argv[++n];
	} else if (!strcmp(argv[n], "-i") && n+1<argc) {
	    if (fatName) {
		fprintf(stderr, "Error: only one FAT image file allowed, "
			"%s already specified\n", fatName);
		err = 1;
	    }
	    fatName = argv[++n];
	} else {
	    err = 1;
	}
    }
    if (err || (fatName && voiceFiles)) {
	fprintf(stderr,
		"pcflash 1.1 (" __DATE__ ")\n"
		"Usage: %s [-p <port>] [-s <speed>] [-l <firmwarefile>]\n"
		"      [-i <fatimage>]\n"
		"      [-f <voicemailfile>]*    (Also sets -r 32 and -noinvert)\n"
		"      [-r <reservedsectors>]   Default: 64\n"
		"      [-noinvert]              Do not invert audio image data\n"
		"Example: pcflash -p 1 -s 115200 -r 64 -l spiall.spi -i content.raw\n",
		argv[0]);
	exit(EXIT_FAILURE);
    }

    if (voiceFiles) {
	FILE *fp;
	tmpName = "voice.img"; //tmpnam("voice");
	if ((fp = fopen(tmpName, "wb"))) {
	    int i;
	    for (i=0;i<voiceFiles;i++) {
		if (AddVoiceFile(fp, voiceFile[i])) {
		    fprintf(stderr, "Could not add %s\n", voiceFile[i]);
		    exit(EXIT_FAILURE);
		}
	    }
	    fclose(fp);
	} else {
	    fprintf(stderr, "Could not open %s for writing\n", tmpName);
	    exit(EXIT_FAILURE);
	}
    }


    if (RS_InitComm(port)) {
	fprintf(stderr, "Could not open serial!\n");
	exit(-1);
    }
    RS_SetCommSpeed(115200);

    printf("Data is %s, reserved blocks %d\n", useInvertedData ? "inverted" : "non-inverted", reservedSectors);
    fprintf(stderr, "Trying to connect..\n");
    /*Wake up monitor*/
    RS_SendData("\xef", 1);
    {
	volatile int breakVar = 1;
	int tries = 8;
	RS_SetBreakIntVar(&breakVar);
	while (1) {
	    unsigned char d;
	    int c;
	    if (!RS_ReceiveData((char *)&d, 1)) {
		c = d;
		fprintf(stderr, "%02x ", c);
		if (c == 0xa5) {
		    fprintf(stderr, "VS10XX\n");
		    //break; /*VS1053*/
		}
		if (c == 0xfe) {
		    /* VS1000 */
		    fprintf(stderr, "VS1000\n");
		    break;
		}
	    } else if (--tries > 0) {
		fprintf(stderr, "Trying again.\n");
		RS_SendData("\xef", 1);
	    } else {
		fprintf(stderr, "Timeout, giving up. Check power and cables.\n");
		RS_CloseComm();
		return 10;
	    }
	}
	RS_SetBreakIntVar(NULL);
    }

    {
	putch(0x72);
	putword(0); //len-1
	putword(0);          //addr hi
	putword(6720+5);   //addr lo player.randomOn
	putch(0<<2); //0=write, 0<<2 = XMEM, 1<<2 = YMEM
	putword(baudRate / 10);
	fprintf(stderr, "Setting up %ld bps\n", (long)baudRate);
    }

    /* load+start software */
    {
	unsigned short *p = &autoprog[0];
	unsigned short head[3];
	int j;
	while (1) {
	    memcpy(head, p, sizeof(head));
	    p += 3;

	    if (head[0] == 0x8000U) {
		//printf("I\n");

		putch(0x72); /*'r'*/
		putword(head[1]>>1); //len-1
		putword(0);          //addr hi
		putword(head[2]);   //addr lo
		putch(2<<2); //0=write, 2<<2 = IMEM

		head[1] += 1;
		while (head[1] > 0) {
		    register int words = head[1];
		    if (words > 256)
			words = 256;
		    head[1] -= words;

		    for (j=0;j<words;j++) {
			putword(*p++);
		    }
		}
	    } else if (head[0] == 0x8001U || head[0] == 0x8002U) {
		//printf(head[0] == 0x8001U ? "X\n" : "Y\n");

		putch(0x72);
		putword(head[1]);
		putword(0);
		putword(head[2]);
		putch((head[0]-1)<<2); //0=write, 0<<2 = XMEM, 1<<2 = YMEM

		head[1] += 1;
		while (head[1] > 0) {
		    register int words = head[1];
		    if (words > 256)
			words = 256;
		    head[1] -= words;
		    for (j=0;j<words;j++) {
			putword(*p++);
		    }
		}
		if (head[0] == 0x8002U) {
		    /*Because of the bug..skip a word*/
		    p+=1;
		}

	    } else if (head[0] == 0x8003U) {
		printf("executing VS1000 code\n");

		putch('e'); /*exec*/
		putword(0);
		putword(head[2]);
		putword(0x0010); /*emuarea*/
		putword(0x0000); /*nop*/
		putword(0x0024);
		break;
	    } else {
		break;
	    }
	}
    }
    if (baudRate != 115200) {
	RS_Wait();
	RS_Drain();
	RS_Wait();
	RS_SetCommSpeed(baudRate);
	RS_Wait();
	RS_Drain();
	RS_Wait();
    }

    if (fatName) {
	fprintf(stderr, "Programming %s at sector %d\n",
		fatName, reservedSectors);
	if (SendFile(fatName, reservedSectors, 0, inName ?0:1)) {
	    fprintf(stderr, "Programming failed\n");
	    RS_CloseComm();
	    return 10;
	}
    } else if (voiceFiles) {
	fprintf(stderr, "Programming %s at sector %d\n",
		tmpName, reservedSectors);
	if (SendFile(tmpName, reservedSectors, 0, inName ?0:1)) {
	    fprintf(stderr, "Programming failed\n");
	    RS_CloseComm();
	    return 10;
	}
    }
    if (inName) {
	fprintf(stderr, "Programming %s\n", inName);
	if (SendFile(inName, 0, (fatName || voiceFiles) ? 1 : 0, 1/*last*/)) {
	    fprintf(stderr, "Programming failed\n");
	    RS_CloseComm();
	    return 10;
	}
    }
    fprintf(stderr, "Programming Finished Successfully\n");
    RS_CloseComm();
    etime = time(NULL);
    fprintf(stderr, "%ld blocks programmed in %d seconds, %5.2fkB/sec\n",
	    sentblocks, (int)(etime-stime), sentblocks*0.5/(etime-stime));
    return 0;
}
